package ai.nextbillion.octant;

import ai.nextbillion.octant.rest.ApiCallback;
import ai.nextbillion.octant.rest.ApiClient;
import ai.nextbillion.octant.rest.ApiException;
import ai.nextbillion.octant.rest.api.OctantApi;
import ai.nextbillion.octant.rest.model.QueryAggregationRequest;
import ai.nextbillion.octant.rest.model.QueryAggregationResponse;
import ai.nextbillion.octant.rest.model.QueryCoordinateRequest;
import ai.nextbillion.octant.rest.model.QueryCoordinateResponse;
import ai.nextbillion.octant.rest.model.Rank;
import ai.nextbillion.octant.service.OctantLocalConnection;

import android.Manifest;
import android.app.Activity;
import android.content.Context;
import android.text.TextUtils;

import androidx.annotation.NonNull;

import com.google.gson.Gson;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Map;

import io.flutter.Log;
import io.flutter.embedding.engine.plugins.FlutterPlugin;
import io.flutter.embedding.engine.plugins.activity.ActivityAware;
import io.flutter.embedding.engine.plugins.activity.ActivityPluginBinding;
import io.flutter.plugin.common.MethodCall;
import io.flutter.plugin.common.MethodChannel;
import io.flutter.plugin.common.MethodChannel.MethodCallHandler;
import io.flutter.plugin.common.MethodChannel.Result;
import pub.devrel.easypermissions.EasyPermissions;
import pub.devrel.easypermissions.PermissionRequest;

/**
 * OctantPlugin
 */
public class OctantPlugin implements FlutterPlugin, MethodCallHandler, ActivityAware {

    private static final String METHOD_CHANNEL_INIT = "init";
    private static final String METHOD_CHANNEL_UPDATE_USER_DETAILS = "updateUserDetails";
    private static final String METHOD_CHANNEL_START_SERVICE = "startService";
    private static final String METHOD_CHANNEL_START_TRACING = "startTracing";
    private static final String METHOD_CHANNEL_STOP_TRACING = "stopTracing";
    private static final String METHOD_CHANNEL_STOP_SERVICE = "stopService";
    private static final String METHOD_CHANNEL_CONFIGURE_INTERVALS = "configureIntervals";
    private static final String METHOD_CHANNEL_QUERY_COORDINATE = "queryUserCoordinate";
    private static final String METHOD_CHANNEL_QUERY_RANKING = "queryRanking";
    private static final String METHOD_CHANNEL_ENABLE_DEBUG = "enableDebug";

    private static final String ARGUMENT_BASE_URL = "baseUrl";
    private static final String ARGUMENT_TEAM_CODE = "teamCode";
    private static final String ARGUMENT_START = "start";
    private static final String ARGUMENT_END = "end";
    private static final String ARGUMENT_LIMIT = "limit";
    private static final String ARGUMENT_REMARK_CODE = "remarkCode";
    private static final String ARGUMENT_USER_CODE = "userCode";
    private static final String ARGUMENT_USER_CODES = "userCodes";
    private static final String ARGUMENT_GATHERING_INTERVAL = "gatheringInterval";
    private static final String ARGUMENT_UPLOADING_INTERVAL = "uploadingInterval";

    private static final String ARGUMENT_NOTIFICATION_ICON_ID = "notificationIconId";
    private static final String ARGUMENT_NOTIFICATION_TITLE = "notificationTitle";
    private static final String ARGUMENT_NOTIFICATION_CONTENT = "notificationContent";

    private static final String ARGUMENT_ENABLE_DEBUG = "enableDebug";

    public static boolean debugEnabled = false;

    private MethodChannel channel;
    private Activity activity;
    private Context context;
    private OctantLocalConnection localServiceConn;
    private String octantBaseUrl;
    private ApiClient apiClient;

    public OctantPlugin() {
        Logger.e("OctantPlugin", "Constructor");
    }

    @Override
    public void onAttachedToEngine(@NonNull FlutterPluginBinding flutterPluginBinding) {
        Logger.e("OctantPlugin", "onAttachedToEngine");
        channel = new MethodChannel(flutterPluginBinding.getBinaryMessenger(), "octant");
        channel.setMethodCallHandler(this);
        context = flutterPluginBinding.getApplicationContext();
        Logger.e("OctantPlugin", "localServiceConn == null?" + (localServiceConn == null));
        apiClient = new ApiClient();
        localServiceConn = new OctantLocalConnection(context);
    }

    private void initApiClient() {
        Log.e("OctantPlugin", "initApiClient:" + octantBaseUrl);
        if (TextUtils.isEmpty(octantBaseUrl)) {
            int resId = context.getResources().getIdentifier("octant_base_url", "string", context.getPackageName());
            Log.e("OctantPlugin", "initApiClient resId:" + resId);
            octantBaseUrl = resId > 0 ? context.getString(resId) : null;
            Log.e("OctantPlugin", "initApiClient octantBaseUrl:" + octantBaseUrl);
        }

//        if (TextUtils.isEmpty(octantBaseUrl)) {
//            throw new RuntimeException("Have you configured Octant base url?");
//        }
//
        if (!TextUtils.isEmpty(octantBaseUrl)) {
            apiClient.setBasePath(octantBaseUrl);
        }


    }

    @Override
    public void onMethodCall(@NonNull MethodCall call, @NonNull Result result) {
        switch (call.method) {
            case METHOD_CHANNEL_INIT:
                String userId = call.argument(ARGUMENT_USER_CODE);
                String teamCode = call.argument(ARGUMENT_TEAM_CODE);
                octantBaseUrl = call.argument(ARGUMENT_BASE_URL);
                initApiClient();
                localServiceConn.init(userId, teamCode);
                break;
            case METHOD_CHANNEL_UPDATE_USER_DETAILS:
                userId = call.argument(ARGUMENT_USER_CODE);
                teamCode = call.argument(ARGUMENT_TEAM_CODE);
                result.success(localServiceConn.updateUser(userId, teamCode));
                break;
            case METHOD_CHANNEL_START_SERVICE:
                String notificationDrawable = call.argument(ARGUMENT_NOTIFICATION_ICON_ID);
                String notificationTitle = call.argument(ARGUMENT_NOTIFICATION_TITLE);
                String notificationContent = call.argument(ARGUMENT_NOTIFICATION_CONTENT);
                localServiceConn.startService(result, notificationDrawable, notificationContent, notificationTitle);
                break;
            case METHOD_CHANNEL_START_TRACING:
                result.success(localServiceConn.startTracing());
                break;
            case METHOD_CHANNEL_STOP_TRACING:
                result.success(localServiceConn.stopTracing());
                break;
            case METHOD_CHANNEL_STOP_SERVICE:
                result.success(localServiceConn.stopService());
                break;
            case METHOD_CHANNEL_CONFIGURE_INTERVALS:
                Integer gathering = call.argument(ARGUMENT_GATHERING_INTERVAL);
                Integer uploading = call.argument(ARGUMENT_UPLOADING_INTERVAL);
                long gatheringInterval = gathering != null ? gathering * 1000 : 5 * 1000L;
                long uploadingInterval = uploading != null ? uploading * 1000 : 20 * 1000L;
                result.success(localServiceConn.configureIntervals(gatheringInterval, uploadingInterval));
                break;
            case METHOD_CHANNEL_QUERY_COORDINATE:
                try {
                    queryUserCoordinate(call, result);
                } catch (ApiException e) {
                    e.printStackTrace();
                }
                break;
            case METHOD_CHANNEL_QUERY_RANKING:
                try {
                    queryRanking(call, result);
                } catch (ApiException e) {
                    e.printStackTrace();
                }
                break;
            case METHOD_CHANNEL_ENABLE_DEBUG:
                Boolean enableDebug = call.argument(ARGUMENT_ENABLE_DEBUG);
                debugEnabled = enableDebug != null ? enableDebug : false;
                break;
            default:
                Logger.e("notImplemented", "method:" + call.method);
                result.notImplemented();
                break;
        }
    }

    @Override
    public void onDetachedFromEngine(@NonNull FlutterPluginBinding binding) {
        Logger.e("OctantPlugin", "onDetachedFromEngine");
        localServiceConn.unbindService();
        channel.setMethodCallHandler(null);
        context = null;
    }

    ///////////////////////////////////////////////////////////////////////////
    // Http APIs
    ///////////////////////////////////////////////////////////////////////////

    private boolean argumentsNullCheck(String... args) {
        for (String arg : args) {
            if (TextUtils.isEmpty(arg)) {
                Logger.e("argumentsNullCheck", "empty:" + arg);
                return false;
            }
        }
        return true;
    }

    private void queryRanking(@NonNull MethodCall call, @NonNull Result callResult) throws ApiException {

        String start = call.argument(ARGUMENT_START);
        String end = call.argument(ARGUMENT_END);
        String limit = call.argument(ARGUMENT_LIMIT);
        String remarkCode = call.argument(ARGUMENT_REMARK_CODE);
        String userCodes = call.argument(ARGUMENT_USER_CODES);
        String teamCode = localServiceConn.getTeamCode();


        List<String> userCodeList = null;
        if (!TextUtils.isEmpty(userCodes)) {
            assert userCodes != null;
            String[] codes = userCodes.split(",");
            userCodeList = Arrays.asList(codes);
        }


        if (argumentsNullCheck(start, end, limit, remarkCode, userCodes, teamCode) && userCodeList != null && !userCodeList.isEmpty()) {
            QueryAggregationRequest request = new QueryAggregationRequest();
            request.end(end)
                    .start(start)
                    .userCode(userCodeList)
                    .limit(limit)
                    .teamCode(teamCode)
                    .remarkCode(remarkCode);

            new OctantApi(apiClient).queryAggregationAsync(request, new ApiCallback<QueryAggregationResponse>() {
                @Override
                public void onFailure(ApiException e, int statusCode, Map<String, List<String>> responseHeaders) {
                    Logger.e("queryRanking", "onFailure:" + statusCode);
                }

                @Override
                public void onSuccess(QueryAggregationResponse result, int statusCode, Map<String, List<String>> responseHeaders) {
                    Logger.e("queryRanking", "onSuccess:" + statusCode);
                    Logger.e("queryRanking", "onSuccess:" + result.toString());
                    List<String> ranks = new ArrayList<>();
                    for (Rank rank : result.getRanks()) {
                        ranks.add(new Gson().toJson(rank));
                    }
                    callResult.success(ranks);
                }

                @Override
                public void onUploadProgress(long bytesWritten, long contentLength, boolean done) {

                }

                @Override
                public void onDownloadProgress(long bytesRead, long contentLength, boolean done) {

                }
            });
        } else {
            callResult.error("Missing Params", "There are missing parameters", null);
        }

    }


    private void queryUserCoordinate(@NonNull MethodCall call, @NonNull Result callResult) throws ApiException {
        QueryCoordinateRequest request = new QueryCoordinateRequest();
        String userCode = localServiceConn.getUserId();
        String teamCode = localServiceConn.getTeamCode();

        String start = call.argument(ARGUMENT_START);
        String end = call.argument(ARGUMENT_END);
        String limit = call.argument(ARGUMENT_LIMIT);
        String remarkCode = call.argument(ARGUMENT_REMARK_CODE);


        if (argumentsNullCheck(start, end, limit, remarkCode, userCode, teamCode)) {
            request.start(start)
                    .end(end)
                    .userCode(userCode)
                    .limit(limit)
                    .teamCode(teamCode)
                    .remarkCode(remarkCode);


            new OctantApi(apiClient).queryCoordinateAsync(request, new ApiCallback<QueryCoordinateResponse>() {
                @Override
                public void onFailure(ApiException e, int statusCode, Map<String, List<String>> responseHeaders) {

                }

                @Override
                public void onSuccess(QueryCoordinateResponse result, int statusCode, Map<String, List<String>> responseHeaders) {

                    callResult.success(new Gson().toJson(result));
                }

                @Override
                public void onUploadProgress(long bytesWritten, long contentLength, boolean done) {

                }

                @Override
                public void onDownloadProgress(long bytesRead, long contentLength, boolean done) {

                }
            });
        } else {
            callResult.error("Missing Params", "There are missing parameters", null);
        }


    }

    ///////////////////////////////////////////////////////////////////////////
    //
    ///////////////////////////////////////////////////////////////////////////

    @Override
    public void onDetachedFromActivity() {
        Logger.e("OctantPlugin", "onDetachedFromActivity");
    }

    @Override
    public void onReattachedToActivityForConfigChanges(ActivityPluginBinding binding) {
        Logger.e("OctantPlugin", "onReattachedToActivityForConfigChanges");
        activity = binding.getActivity();
        localServiceConn.setContext(activity);
    }

    @Override
    public void onAttachedToActivity(ActivityPluginBinding binding) {
        Logger.e("OctantPlugin", "onAttachedToActivity");
        activity = binding.getActivity();
        localServiceConn.setContext(activity);
        requestPermissions();
    }

    @Override
    public void onDetachedFromActivityForConfigChanges() {
        Logger.e("OctantPlugin", "onDetachedFromActivityForConfigChanges");
    }

    ///////////////////////////////////////////////////////////////////////////
    //
    ///////////////////////////////////////////////////////////////////////////

    private final static int REQ_PERMISSIONS = 19;

    private void requestPermissions() {
        String[] permissions;
        if (android.os.Build.VERSION.SDK_INT >= android.os.Build.VERSION_CODES.Q) {
            permissions = new String[]{
                    Manifest.permission.ACCESS_FINE_LOCATION,
                    Manifest.permission.ACCESS_COARSE_LOCATION,
                    Manifest.permission.ACCESS_BACKGROUND_LOCATION};
        } else {
            permissions = new String[]{
                    Manifest.permission.ACCESS_FINE_LOCATION,
                    Manifest.permission.ACCESS_COARSE_LOCATION};
        }

        EasyPermissions.requestPermissions(new PermissionRequest.Builder(activity, REQ_PERMISSIONS, permissions)
                .setRationale("EagleVision requires location access")
                .setNegativeButtonText("No")
                .setPositiveButtonText("OK")
                .build());
    }
}
