package ai.nextbillion.octant.service;

import android.app.Notification;
import android.app.NotificationChannel;
import android.app.NotificationManager;
import android.app.PendingIntent;
import android.app.Service;
import android.content.Context;
import android.content.Intent;
import android.content.pm.PackageManager;
import android.location.Location;
import android.os.Binder;
import android.os.Build;
import android.os.IBinder;
import android.text.TextUtils;

import androidx.annotation.NonNull;
import androidx.core.app.NotificationCompat;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Timer;
import java.util.TimerTask;

import ai.nextbillion.octant.Logger;
import ai.nextbillion.octant.R;
import ai.nextbillion.octant.db.LocationEntity;
import ai.nextbillion.octant.db.LocationEntity_;
import ai.nextbillion.octant.db.ObjectBox;
import ai.nextbillion.octant.engine.LocationEngine;
import ai.nextbillion.octant.engine.LocationEngineCallback;
import ai.nextbillion.octant.engine.LocationEngineProvider;
import ai.nextbillion.octant.engine.LocationEngineRequest;
import ai.nextbillion.octant.engine.LocationEngineResult;
import ai.nextbillion.octant.rest.ApiCallback;
import ai.nextbillion.octant.rest.ApiException;
import ai.nextbillion.octant.rest.api.OctantApi;
import ai.nextbillion.octant.rest.model.Coordinate;
import ai.nextbillion.octant.rest.model.UploadPositionRequest;
import ai.nextbillion.octant.rest.model.UploadPositionResponse;
import ai.nextbillion.octant.rest.model.UserCoords;
import io.objectbox.Box;
import io.objectbox.query.Query;

public class OctantLocalService extends Service {

    public static boolean isServiceRunning = false;

    private final IBinder localBinder = new LocalBinder();
    private final OctantApi api = new OctantApi();

    private Box<LocationEntity> locationBox;
    private Query<LocationEntity> locationQuery;

    private String userId;
    private String teamCode;

    private boolean isTraceRunning = false;


    private String notificationDrawable;
    private String notificationContent;
    private String notificationTitle;

    @Override
    public void onCreate() {
        Logger.e("OctantLocalService", "onCreate");
        super.onCreate();
        isServiceRunning = true;
    }

    @Override
    public void onRebind(Intent intent) {
        super.onRebind(intent);
        Logger.e("OctantLocalService", "onRebind");
    }

    @Override
    public IBinder onBind(Intent intent) {
        Logger.e("OctantLocalService", "onBind");
        String notificationDrawable = intent.getStringExtra("notificationDrawable");
        String notificationContent = intent.getStringExtra("notificationContent");
        String notificationTitle = intent.getStringExtra("notificationTitle");
        this.notificationDrawable = TextUtils.isEmpty(notificationDrawable) ? this.notificationDrawable : notificationDrawable;
        this.notificationTitle = TextUtils.isEmpty(notificationTitle) ? this.notificationTitle : notificationTitle;
        this.notificationContent = TextUtils.isEmpty(notificationContent) ? this.notificationContent : notificationContent;
        Logger.e("OctantLocalService", "notificationDrawable:" + notificationDrawable);
        startForeground(REPORTING_NOTIFICATION_ID, buildNotification(this.notificationDrawable, this.notificationContent, this.notificationTitle));
        openBox();
        return localBinder;
    }

    @Override
    public int onStartCommand(Intent intent, int flags, int startId) {
        Logger.e("OctantLocalService", "onStartCommand flags:" + flags);
        Logger.e("OctantLocalService", "onStartCommand startId:" + startId);
        createNotificationChannel(this);
        return START_STICKY;
    }

    @Override
    public boolean onUnbind(Intent intent) {
        Logger.e("OctantLocalService", "onUnbind");
        return super.onUnbind(intent);
    }

    @Override
    public void onTaskRemoved(Intent rootIntent) {
        super.onTaskRemoved(rootIntent);
        Logger.e("OctantLocalService", "onTaskRemoved");
    }

    @Override
    public void onDestroy() {
        Logger.e("OctantLocalService", "onDestroy");
        closeBox();
        stopForeground(true);
        if (isTraceRunning) {
            stopTracing();
        }
        isServiceRunning = false;
        super.onDestroy();
    }

    ///////////////////////////////////////////////////////////////////////////
    // Notification
    ///////////////////////////////////////////////////////////////////////////

    public static final int REPORTING_NOTIFICATION_ID = 6501;
    public static final String REPORTING_NOTIFICATION_CHANNEL = "REPORTING_NOTIFICATION_CHANNEL";

    public void startTracing() {
        Logger.e("OctantLocalService", "startLocationReporting");
        if (!isTraceRunning) {
            activeLocationEngine(currentGatheringInterval);
            startTimer(currentUploadingInterval);
            isTraceRunning = true;
        }
    }

    public void stopTracing() {
        cancelTimer();
        deactivateLocationEngine();
        isTraceRunning = false;
    }

    public void setUser(String userId, String teamCode) {
        this.userId = userId;
        this.teamCode = teamCode;
    }

    public class LocalBinder extends Binder {
        public OctantLocalService getService() {
            return OctantLocalService.this;
        }
    }

    Notification buildNotification(String drawableName, String text, String title) {
        PendingIntent pendingOpenIntent = createPendingOpenIntent(this);
        NotificationCompat.Builder builder = new NotificationCompat.Builder(getApplicationContext(), REPORTING_NOTIFICATION_CHANNEL)
                .setContentIntent(pendingOpenIntent)
                .setPriority(NotificationCompat.PRIORITY_MAX)
                .setCategory(NotificationCompat.CATEGORY_SERVICE)
                .setSmallIcon(getDrawableId(drawableName))
                .setContentTitle(title)
                .setOngoing(true);

        if (!TextUtils.isEmpty(text)) {
            builder.setContentText(text);
        }

        if (!TextUtils.isEmpty(title)) {
            builder.setContentTitle(title);
        }
        return builder.build();
    }

    private int getDrawableId(String name) {
        int resId = 0;
        if (!TextUtils.isEmpty(name)) {
            int resIdFromMipmap = getResources().getIdentifier(name, "mipmap", getPackageName());
            int resIdFromDrawable = getResources().getIdentifier(name, "drawable", getPackageName());
            Logger.e("OctantLocalService", "notificationDrawable resIdFromMipmap:" + resIdFromMipmap);
            Logger.e("OctantLocalService", "notificationDrawable resIdFromDrawable:" + resIdFromDrawable);
            resId = resIdFromDrawable;

            if (resIdFromMipmap > 0) {
                resId = resIdFromMipmap;
            }
        }
        return resId;
    }

    private void createNotificationChannel(Context context) {
        if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.O) {
            NotificationManager notificationManager = (NotificationManager) context.getSystemService(Context.NOTIFICATION_SERVICE);
            NotificationChannel notificationChannel = new NotificationChannel(
                    REPORTING_NOTIFICATION_CHANNEL, context.getString(R.string.channel_name),
                    NotificationManager.IMPORTANCE_LOW);
            notificationManager.createNotificationChannel(notificationChannel);
        }
    }

    private PendingIntent createPendingOpenIntent(Context context) {
        PackageManager pm = context.getPackageManager();
        Intent intent = pm.getLaunchIntentForPackage(context.getPackageName());
        intent.setPackage(null);
        return PendingIntent.getActivity(context, 0, intent, 0);
    }

    ///////////////////////////////////////////////////////////////////////////
    // Timer
    ///////////////////////////////////////////////////////////////////////////

    private Timer timer;

    private void startTimer(long interval) {
        if (timer != null) {
            timer.cancel();
            timer.purge();
        }
        timer = new Timer();
        timer.scheduleAtFixedRate(new OctantLocalService.UploadingTimerTask(), interval, interval);
    }

    private void cancelTimer() {
        Logger.e("Local Service", "cancelTimer");
        if (timer != null) {
            timer.cancel();
            timer.purge();
        }
    }

    final class UploadingTimerTask extends TimerTask {

        @Override
        public void run() {
            uploadLocations();
        }
    }


    ///////////////////////////////////////////////////////////////////////////
    //
    ///////////////////////////////////////////////////////////////////////////
    static final long DEFAULT_GATHERING_INTERVAL_MILLIS = 5 * 1000;
    static final long MAX_GATHERING_INTERVAL_MILLIS = 60 * 60 * 1000;
    static final long DEFAULT_UPLOADING_INTERVAL_MILLIS = 25 * 1000;

//    List<Location> cachedLocations = Collections.synchronizedList(new ArrayList<Location>());

    private void uploadLocations() {
//        List<Location> locationsToBeUploaded = new ArrayList<>(cachedLocations);
//        cachedLocations.clear();

        List<LocationEntity> locationsToBeUploaded = findAllFromDB();
        Logger.e("uploadLocations", "uploading...size:" + locationsToBeUploaded.size());
        Logger.e("uploadLocations", "uploading...:" + locationsToBeUploaded);

        UploadPositionRequest request = new UploadPositionRequest();
        List<Coordinate> coords = new ArrayList<>();
        for (LocationEntity location : locationsToBeUploaded) {
            coords.add(new Coordinate()
                    .lat(location.latitude)
                    .lng(location.longitude)
                    .ts(String.valueOf(location.timestamp)));
        }

        if (!coords.isEmpty()) {
            request
                    .coords(new UserCoords()
                            .coords(coords)
                            .teamCode(teamCode)
                            .userCode(userId))
                    .updated("0");


            try {
                api.uploadCoordinateAsync(request, new ApiCallback<UploadPositionResponse>() {
                    @Override
                    public void onFailure(ApiException e, int statusCode, Map<String, List<String>> responseHeaders) {
                        //Do nothing
                        Logger.e("uploadLocations", "uploading failed");
                    }

                    @Override
                    public void onSuccess(UploadPositionResponse result, int statusCode, Map<String, List<String>> responseHeaders) {
                        removeUploadedLocation(locationsToBeUploaded);
                        Logger.e("uploadLocations", "uploading done");
                    }

                    @Override
                    public void onUploadProgress(long bytesWritten, long contentLength, boolean done) {
                        Logger.e("uploadLocations", "onUploadProgress:" + contentLength);
                    }

                    @Override
                    public void onDownloadProgress(long bytesRead, long contentLength, boolean done) {

                    }
                });
            } catch (ApiException e) {
                e.printStackTrace();
            }
        }
    }

    private LocationEngine locationEngine;


    private long currentGatheringInterval = DEFAULT_GATHERING_INTERVAL_MILLIS;
    private long currentUploadingInterval = DEFAULT_UPLOADING_INTERVAL_MILLIS;

    public void configureIntervals(long gatheringInterval, long uploadingInterval) {
        if (currentGatheringInterval != gatheringInterval) {
            Logger.e("configureIntervals", "input gathering:" + gatheringInterval);
            Logger.e("configureIntervals", "input uploading:" + uploadingInterval);

            currentGatheringInterval = roundingGatheringInterval(gatheringInterval);
            Logger.e("configureIntervals", "rounded gathering:" + currentGatheringInterval);

            if (isTraceRunning) {
                activeLocationEngine(currentGatheringInterval);
            }
        }

        if (currentUploadingInterval != uploadingInterval) {
            currentUploadingInterval = roundingUploadingInterval(uploadingInterval);
            Logger.e("configureIntervals", "rounded uploading:" + currentUploadingInterval);
            if (isTraceRunning) {
                startTimer(currentUploadingInterval);
            }
        }
    }

    private long roundingGatheringInterval(long interval) {
        if (interval <= DEFAULT_GATHERING_INTERVAL_MILLIS) {
            return DEFAULT_GATHERING_INTERVAL_MILLIS;
        }

        if (interval > MAX_GATHERING_INTERVAL_MILLIS) {
            return MAX_GATHERING_INTERVAL_MILLIS;
        }

        return interval / 5000 * 5000;
    }

    private long roundingUploadingInterval(long interval) {
        if (interval < currentGatheringInterval * 4) {
            return currentGatheringInterval * 4;
        }

        if (interval > currentGatheringInterval * 10) {
            return currentGatheringInterval * 10;
        }

        return interval / currentGatheringInterval * currentGatheringInterval;
    }

    private void deactivateLocationEngine() {
        Logger.e("Local Service", "deactivateLocationEngine");
        if (locationEngine != null) {
            locationEngine.removeLocationUpdates(locationEngineCallback);
        }
    }

    private void activeLocationEngine(long interval) {
        if (locationEngine == null) {
            locationEngine = LocationEngineProvider.getBestLocationEngine(getApplicationContext());
        } else {
            locationEngine.removeLocationUpdates(locationEngineCallback);
        }
        locationEngine.requestLocationUpdates(buildEngineRequest(interval), locationEngineCallback, null);
        locationEngine.getLastLocation(locationEngineCallback);
    }


    private final LocationEngineCallback<LocationEngineResult> locationEngineCallback = new LocationEngineCallback<LocationEngineResult>() {
        @Override
        public void onSuccess(LocationEngineResult result) {
            Location location = result.getLastLocation();
            if (location != null) {
                Logger.e("LocationEngine", "onLocationFound:" + location.toString());
//                cachedLocations.add(location);
                putLocation(location);
            }

        }

        @Override
        public void onFailure(@NonNull Exception exception) {

        }
    };

    private LocationEngineRequest buildEngineRequest(long interval) {
        return new LocationEngineRequest.Builder(interval)
                .setFastestInterval(interval)
                .setPriority(LocationEngineRequest.PRIORITY_HIGH_ACCURACY)
                .build();
    }

    ///////////////////////////////////////////////////////////////////////////
    //  Database
    ///////////////////////////////////////////////////////////////////////////

    private List<LocationEntity> findAllFromDB() {
        List<LocationEntity> locationEntityList = locationQuery.find();
        return locationEntityList;
    }

    private void putLocation(Location location) {
        LocationEntity entity = new LocationEntity();
        entity.id = location.getTime() % 1000;
        entity.latitude = location.getLatitude();
        entity.longitude = location.getLongitude();
        entity.timestamp = location.getTime();
        locationBox.put(entity);
    }

    private void removeUploadedLocation(List<LocationEntity> uploaded) {
        locationBox.remove(uploaded);
    }

    private void openBox() {
        if (ObjectBox.get() != null && !ObjectBox.get().isClosed()) {
            ObjectBox.get().close();
        }
        ObjectBox.init(this.getApplicationContext());
        locationBox = ObjectBox.get().boxFor(LocationEntity.class);
        locationQuery = locationBox.query().order(LocationEntity_.timestamp).build();
    }

    private void closeBox() {
        if (!ObjectBox.get().isClosed()) {
            ObjectBox.get().close();
        }
    }

    ///////////////////////////////////////////////////////////////////////////
    // Stop Service
    ///////////////////////////////////////////////////////////////////////////

    public void stopService() {
        stopForeground(true);
    }

}